/**********************************************************************
WiLight - Web control interface for Neopixel ring.
          Part of the VoiceBox project.
Target: ESP-01 esp8266 module
Wiring: see https://hackaday.io/project/160052-voicebox/log/151310-hooking-up-the-esp-01

TM 180822
***********************************************************************/
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <Adafruit_NeoPixel.h>

#define SSID      "YOUR_SSID"
#define PASSWORD  "YOUR_PASSWORD"

MDNSResponder mdns;

ESP8266WebServer server(80);
#define BIN2RAD       2*PI/256.0
#define NEOPIXEL_PIN  2 // GPIO2
#define PIXELS       12 // number of neopixels in strand

// Parameter 1 = number of pixels in strip
// Parameter 2 = pin number (most are valid)
// Parameter 3 = pixel type flags, add together as needed:
//   NEO_KHZ800  800 KHz bitstream (most NeoPixel products w/WS2812 LEDs)
//   NEO_KHZ400  400 KHz (classic 'v1' (not v2) FLORA pixels, WS2811 drivers)
//   NEO_GRB     Pixels are wired for GRB bitstream (most NeoPixel products)
//   NEO_RGB     Pixels are wired for RGB bitstream (v1 FLORA pixels, not v2)
Adafruit_NeoPixel strip = Adafruit_NeoPixel(PIXELS, NEOPIXEL_PIN, NEO_GRB + NEO_KHZ800);

// theaterChase() modes
#define THEATER_RANDOM  0
#define THEATER_COLOR   1
#define THEATER_FLAMES  2
#define THEATER_WAVES   3

#define PATTERN_NONE    0
#define PATTERN_SPIN    1
#define PATTERN_PULSE   2
#define PATTERN_CHASE   3
#define PATTERN_RAINBOW 4
#define PATTERN_WAVES   5
#define PATTERN_FLAMES  6
#define PATTERN_LARSON  7
#define PATTERN_SOLID   8
#define NUM_PATTERNS    9

String pattern_names[] = {"none", "spin", "pulse", "chase", "rainbow", "waves", "flames", "larson", "solid"};

int pattern = PATTERN_NONE;
int cycle = 0; // cycles through 0..255, used for cyclical variation in patterns
#define BLACK   strip.Color(0x00,0x00,0x00)
#define BLUE    strip.Color(0x00,0x00,0xFF)
#define GREEN   strip.Color(0x00,0xFF,0x00)
#define CYAN    strip.Color(0x00,0xFF,0xFF)
#define RED     strip.Color(0xFF,0x00,0x00)
#define MAGENTA strip.Color(0xFF,0x00,0xFF)
#define ORANGE  strip.Color(0xFF,0x45,0x00)
#define YELLOW  strip.Color(0xFF,0xFF,0x00)
#define WHITE   strip.Color(0xFF,0xFF,0xFF)

#define DEFAULT_COLOUR "#E1D34D"

uint32_t custom_colour = BLACK; // initialized to DEFAULT in setup
String hex_colour = "#000000";
double custom_hsv[] = {0.0, 0.0, 0.0};

bool use_custom_colour = false; // use custom colour or default rainbow where available

uint32_t pixel_colour = BLACK;

// ************************ PIXEL UTILITIES ***************************
void setAll(uint32_t colour)
{
  for (int i=0; i < strip.numPixels(); i++)
    strip.setPixelColor(i, colour);
  strip.show();
}

// ************* RAINBOW *************
void rainbow(int cycle, uint8_t wait)
{
  float spacing = 360.0 / 16.0;
  float hue = (millis() / 10) % 360;
  for(int x=0; x <= strip.numPixels()-1; x++)
  {
    float offset = x * spacing;
    double h = (int(hue + offset) % 360) / 360.0;
    strip.setPixelColor(x, hsvToRgb(h, 1.0, 1.0));
  }
  strip.show();
  delay(wait);
}

// ************* THEATER CHASE *************
// Theatre-style crawling lights,
// if mode THEATER_COLOR, all of color c
// if mode THEATER_RANDOM, or c is 0, use a random color
// if THEATER_FLAMES, use red and orange
// if THEATER_WAVES, use blue and cyan
void theaterChase(int cycle, uint8_t m, uint32_t c, uint8_t wait)
{
  for (int q=0; q < 3; q++)
  {
    for (int i=0; i < strip.numPixels(); i=i+3)
    {
      uint32_t colour = c;
      if (m==THEATER_FLAMES)
        colour = (random(0,3) > 0 ? RED : ORANGE);
      else if (m==THEATER_WAVES)
        colour = (random(0,3) > 0 ? BLUE : CYAN);
      else if (m==THEATER_RANDOM || c==0) 
        colour = color8(random(1,7));
      strip.setPixelColor(i+q, colour); // every third pixel on
    }
    strip.show();
     
    delay(wait);
     
    for (int i=0; i < strip.numPixels(); i=i+3) 
      strip.setPixelColor(i+q, 0);  // every third pixel off
  }
}

// ************* PULSE *************
void pulse(uint8_t wait, int cycle)
{
  float FALLOFF = 5.9 * cycle/256.0;
  float SCAN_SPEED = 1.5;
  double offset = 0.5;
  double phase = cycle /256.0;

  // Use custom colour or use offset to pick the right colour from the hue wheel
  double hue = int(round((use_custom_colour ? custom_hsv[0] : phase) * 360));

  // Maximum distance from lead pixel
  int max_distance = strip.numPixels() -1;

  // Scale offset to [0..max_distance]
  // offset = int(round(offset * max_distance));
  offset = offset * max_distance;

  for(int x=0; x<strip.numPixels(); x++)
  {
    double distance = offset -x;
    if (distance < 0) distance = -distance;
    distance += 2;
    double val = max_distance - (distance * FALLOFF);
    val /= double(max_distance);    // Scale to [0.0..1.0]
    val = min(max(val,0.0), 1.0);   // Ditch negative values

    double xhue = hue;              // Grab hue for this pixel
    xhue += (1-val) * 10;           // Use the val offset to give a slight colour trail variation
    xhue = int(xhue) % 360;         // Clamp to 0-359
    xhue /= 360.0;                  // Scale to [0.0..1.0]
    strip.setPixelColor(x, hsvToRgb(xhue, 1.0, val/4));
  }
  strip.show();      
  delay(wait);
}


// ************* SPIN *************
// Spinning comet trail
void spin(uint8_t wait, int cycle)
{
  float FALLOFF = 2.9;
  double offset = cycle/256.0;

  // Use custom colour or use offset to pick the right colour from the hue wheel
  double hue = int(round((use_custom_colour ? custom_hsv[0] : offset) * 360));

  // Maximum distance from lead pixel
  int max_distance = strip.numPixels() - 1;

  // Scale offset to [0..max_distance]
  offset = int(round(offset * max_distance));

  for(int x=0; x<strip.numPixels(); x++)
  {
    double sat = 1.0;
    double distance = offset -x;
    if (distance < 0) distance += max_distance;
    double val = max_distance - (distance * FALLOFF);
    val /= double(max_distance);    // Scale to [0.0..1.0]
    val = max(val,0.0);             // Ditch negative values

    double xhue = hue;              // Grab hue for this pixel
    xhue += (1-val) * 10;           // Use the val offset to give a slight colour trail variation
    xhue = int(xhue) % 360;         // Clamp to 0-359
    xhue /= 360.0;                  // Scale to [0.0..1.0]

    strip.setPixelColor(x, hsvToRgb(xhue, sat, val/4));
  }
  strip.show();      
  delay(wait);
}

// ************* LARSON *************
// "Larson lights" Knight Rider/Cylon effect
void larson(uint8_t wait)
{
  float FALLOFF = 2.9;
  float SCAN_SPEED = 1.5;

  double delta = millis()/1000.0;

  // Offset is a sine wave derived from the time delta
  // we use this to animate both the hue and larson scan
  // so they are kept in sync with each other
  double offset = (sin(delta * SCAN_SPEED) + 1) / 2;

  // Use custom colour or use offset to pick the right colour from the hue wheel
  double hue = int(round((use_custom_colour ? custom_hsv[0] : offset) * 360));

  // Maximum distance from lead pixel
  int max_distance = strip.numPixels() - 1;

  // Scale offset to [0..max_distance]
  offset = int(round(offset * max_distance));

  for(int x=0; x<strip.numPixels(); x++)
  {
    double sat = 1.0;

    double val = max_distance - (abs(offset - x) * FALLOFF);
    val /= double(max_distance);    // Scale to [0.0..1.0]
    val = max(val,0.0);             // Ditch negative values

    double xhue = hue;              // Grab hue for this pixel
    xhue += (1-val) * 10;           // Use the val offset to give a slight colour trail variation
    xhue = int(xhue) % 360;         // Clamp to 0-359
    xhue /= 360.0;                  // Scale to [0.0..1.0]

    strip.setPixelColor(x, hsvToRgb(xhue, sat, val/4));
  }
  strip.show();      
  delay(wait);
}

// ************************ COLOUR FUNCTIONS ***************************
// Map a number in the range 0..7 to an RGB color.
// BLACK, BLUE, GREEN, CYAN, RED, MAGENTA, YELLOW, WHITE
uint32_t color8(byte index)
  {return strip.Color(127*(index&4), 127*(index&2), 127*(index&1));}
  
// Map a number in the range 0 to 255 to an RGB color.
//   0.. 85 -> Green -> Yellow  -> Red
//  85..170 -> Red   -> Magenta -> Blue
// 170..255 -> Blue  -> Cyan    -> Green
uint32_t color256(byte n)
{
  if(n < 85)
    return strip.Color(n*3, 255 - n*3, 0);
  else if(n < 170) 
    {n -= 85; return strip.Color(255 - n*3, 0, n*3);}
  else
    {n -= 170; return strip.Color(0, n*3, 255 - n*3);}
}

/**
 * Converts an HSV color value to RGB. Conversion formula
 * adapted from http://en.wikipedia.org/wiki/HSV_color_space.
 * Assumes h, s, and v are contained in the set [0, 1] and
 * returns r, g, and b in the set [0, 255].
 *
 * @param   Number  h       The hue
 * @param   Number  s       The saturation
 * @param   Number  v       The value
 * @return  Array           The RGB representation
 * 
 * From: https://github.com/ratkins/RGBConverter/blob/master/RGBConverter.cpp
 */
uint32_t hsvToRgb(double h, double s, double v)
{
  double r, g, b;

  int i = int(h * 6);
  double f = h * 6 - i;
  double p = v * (1 - s);
  double q = v * (1 - f * s);
  double t = v * (1 - (1 - f) * s);

  switch(i % 6)
  {
    case 0: r = v, g = t, b = p; break;
    case 1: r = q, g = v, b = p; break;
    case 2: r = p, g = v, b = t; break;
    case 3: r = p, g = q, b = v; break;
    case 4: r = t, g = p, b = v; break;
    case 5: r = v, g = p, b = q; break;
  }

  return strip.Color(r*255, g*255, b*255);
}

/**
 * Converts an RGB color value to HSV. Conversion formula
 * adapted from http://en.wikipedia.org/wiki/HSV_color_space.
 * Assumes r, g, and b are contained in the set [0, 255] and
 * returns h, s, and v in the set [0, 1].
 *
 * @param   Number  r       The red color value
 * @param   Number  g       The green color value
 * @param   Number  b       The blue color value
 * @return  Array           The HSV representation
 */
void rgbToHsv(uint32_t rgb, double hsv[])
{
  double rd = (double) ((rgb>>16 & 0x0000FF)/255.0);
  double gd = (double) ((rgb>> 8 & 0x0000FF)/255.0);
  double bd = (double) ((rgb     & 0x0000FF)/255.0);;
  double max = threeway_max(rd, gd, bd), min = threeway_min(rd, gd, bd);
  double h, s, v = max;

  double d = max - min;
  s = max == 0 ? 0 : d / max;

  if (max == min)
    h = 0; // achromatic
  else
  {
    if (max == rd)
      h = (gd - bd) / d + (gd < bd ? 6 : 0);
    else if (max == gd)
      h = (bd - rd) / d + 2;
    else if (max == bd)
      h = (rd - gd) / d + 4;
    h /= 6;
  }

  hsv[0] = h;
  hsv[1] = s;
  hsv[2] = v;
}
double threeway_max(double a, double b, double c) {return max(a, max(b, c));}
double threeway_min(double a, double b, double c) {return min(a, min(b, c));}

// ************************ PAGE HANDLERS ***************************
int hexDigitValue(char ch){return isDigit(ch) ? ch - '0' : 10 + ch - 'A';}
void setColour(String colour_code)
{
  uint32_t int_colour = 0;
  colour_code.toUpperCase();
  if (colour_code.length() != 7) return;
  if (colour_code == hex_colour) return;
  for (int i=1; i<7; i++)
  {
    if (!isHexadecimalDigit(colour_code[i]))
      return;
    else
      int_colour = 16*int_colour + hexDigitValue(colour_code[i]);
  }
  hex_colour = colour_code;
  custom_colour = int_colour;
  rgbToHsv(custom_colour, custom_hsv);
  Serial.print("H: "); Serial.println(custom_hsv[0]);
  Serial.print("S: "); Serial.println(custom_hsv[1]);
  Serial.print("V: "); Serial.println(custom_hsv[2]);
  if (pattern == PATTERN_SOLID) setAll(custom_colour);
  Serial.print("Colour: "); Serial.println(hex_colour);
}

void setPattern(String pattern_name)
{
  if (pattern_name == pattern_names[pattern]) return;
  for (int p=0; p<NUM_PATTERNS; p++)
    if (pattern_name == pattern_names[p]) {pattern = p; break;}
  if (pattern == PATTERN_NONE) setAll(BLACK);
  if (pattern == PATTERN_SOLID) setAll(custom_colour);
  Serial.print("Pattern: "); Serial.println(pattern_names[pattern]);
}

void setUseColour(String colour_to_use)
{
  use_custom_colour = (colour_to_use == "custom");
  Serial.print("use colour: "); Serial.println(colour_to_use);
}

void handleRoot()
{
  if (server.hasArg("hex_colour"))   setColour(server.arg("hex_colour"));
  if (server.hasArg("use_colour"))   setUseColour(server.arg("use_colour"));
  if (server.hasArg("pattern_name")) setPattern(server.arg("pattern_name"));
  
  String html = "<!DOCTYPE html>";
  html += "<head><link href=\"//fonts.googleapis.com/css?family=Raleway:400,300,600\" rel=\"stylesheet\" type=\"text/css\"></head>";
  // margin-left:50px;
  html += "<body style=\"background-color:#757575;font-family:Raleway;color:#FFFFFF;font-size:20px;text-align:center\">";
  html += "<h1 style=\"margin-top:20px;color:#FF8C00\">Web Lamp</h2>";
  html += "<form action=\"/\" method=\"post\">";
  html += "<h2>Select colour</h2>";
  // style=\"margin-left:64px\"
  html += "<input type=\"color\" name=\"hex_colour\" value=\""+hex_colour+"\"><br><br>";
  html += "<h2>Use custom colour or defaults?</h2>";
  html += "<input type=\"radio\" name=\"use_colour\" value=\"custom\""+(use_custom_colour ? String(" checked=\"checked\"") : "")+"/>  Custom<br>";
  html += "<input type=\"radio\" name=\"use_colour\" value=\"default\""+(!use_custom_colour ? String(" checked=\"checked\"") : "")+"/> Default<br>";
  html += "<br>";
  html += "<h2>Select pattern</h2>";
  html += "<select name=\"pattern_name\">";
  for (int p=0; p<NUM_PATTERNS; p++)
    html += "<option value=\""+pattern_names[p]+"\""+(p==pattern ? " selected=\"selected\"" : "")+">"+pattern_names[p]+"</option>";
  html += "</select><br><br><br>";
  html += "<input type=\"submit\" value = \"update\">";
  html += "</form>";
  html += "</div></body></html>";
  
  server.send(200, "text/html", html);
}

void handlePattern()
{
  if (server.hasArg("pattern_name")) setPattern(server.arg("pattern_name"));
  
  String html = "<!DOCTYPE html>";
  
  server.send(200, "text", pattern_names[pattern]);
}

void handleNotFound()
{
  String message = "File Not Found\n\n";
  message += "URI: ";
  message += server.uri();
  message += "\nMethod: ";
  message += (server.method() == HTTP_GET)?"GET":"POST";
  message += "\nArguments: ";
  message += server.args();
  message += "\n";
  for (uint8_t i=0; i<server.args(); i++)
  {
    message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
  }
  server.send(404, "text/plain", message);
}
 

// ************************ SETUP ***************************
void setup(void)
{
  pinMode(NEOPIXEL_PIN, OUTPUT);
  Serial.begin(115200);
  WiFi.begin(SSID, PASSWORD);
  Serial.println("");
  
  strip.begin();
  strip.show(); // Initialize all pixels to 'off'

  // Wait for connection
  while (WiFi.status() != WL_CONNECTED)
  {
    delay(500);
    Serial.print(".");
  }
  Serial.println("");
  Serial.print("Connected to "); Serial.println(SSID);
  Serial.print("IP address: ");  Serial.println(WiFi.localIP());
  
  if (mdns.begin("esp8266", WiFi.localIP()))
    Serial.println("MDNS responder started");
  
  server.on("/", handleRoot);
  server.on("/pattern", handlePattern);
  server.onNotFound(handleNotFound);
  
  server.begin();
  Serial.println("HTTP server started");
  
  setColour(DEFAULT_COLOUR); // set custom colour and associated fields to default
}
 
// ************************ MAIN ***************************
void loop(void)
{
  server.handleClient();
  
  int tempo = 50*(sin(BIN2RAD*cycle)+1.2);
  
  switch(pattern)
  {
    case PATTERN_NONE:
         break;
         
    case PATTERN_SPIN:
         spin(2, cycle);
         break;
      
    case PATTERN_PULSE:
         pulse(10, cycle);
         break;
      
    case PATTERN_CHASE:
         theaterChase(cycle, THEATER_COLOR, custom_colour, 2*tempo);
         break;

    case PATTERN_RAINBOW:
         rainbow(cycle, 20);
         break;
  
    case PATTERN_WAVES:
         theaterChase(cycle, THEATER_WAVES, 0, tempo);
         break;

    case PATTERN_FLAMES: // Gradually speeding flame colors
         theaterChase(cycle, THEATER_FLAMES, 0, tempo);
         break;

    case PATTERN_LARSON:
         larson(100);
         break;
         
    case PATTERN_SOLID:
         break;
            
    default: break; // nothing
  }
  cycle++;
  if(cycle >= 256) cycle = 0;
} 
